print("import: 'smt'")
import smt

print("import: 'smt.surrogate_models'")
import smt.surrogate_models

print("import: 'smt.problems'")
import smt.problems

print("import: 'smt.sampling_methods'")
import smt.sampling_methods

print("import: 'smt.utils'")
import smt.utils

